// Conway's Game of Life for the LILYGO TTGO T-Display dev board.
// Thanks to the following:
// https://github.com/Bodmer/TFT_eSPI - TFT driver library.
// https://github.com/LennartHennigs/Button2 - Arduino Library to simplify working with buttons.
// https://github.com/Xinyuan-LilyGO/TTGO-T-Display - Sample code for the TTGO T-Display board.
#include <TFT_eSPI.h>
#include <SPI.h>
#include "WiFi.h"
#include <Wire.h>
#include <Button2.h>
#include "esp_adc_cal.h"

#ifndef TFT_DISPOFF
#define TFT_DISPOFF 0x28
#endif

#ifndef TFT_SLPIN
#define TFT_SLPIN   0x10
#endif

#define TFT_MOSI            19
#define TFT_SCLK            18
#define TFT_CS              5
#define TFT_DC              16
#define TFT_RST             23

#define TFT_BL          4  // Display backlight control pin
#define ADC_EN          14
#define ADC_PIN         34
#define BUTTON_1        35
#define BUTTON_2        0

#define NATIVE_WIDTH    135
#define NATIVE_HEIGHT   240
#define BIRTH           TFT_GREEN
#define ALIVE           TFT_BLACK
#define DEATH           TFT_RED
#define DEAD            TFT_WHITE
#define PROB            50  // Probability that a cell will start alive

TFT_eSPI tft = TFT_eSPI(NATIVE_WIDTH, NATIVE_HEIGHT); // Invoke custom library
Button2 btn1(BUTTON_1);
Button2 btn2(BUTTON_2);

int scale = 1;
int curWidth = NATIVE_WIDTH;
int curHeight = NATIVE_HEIGHT;

// Our screen buffer. drawBoard will fill this based on the bits in the current and previous
// boards before calling the function to draw to the screen.
unsigned short screenBuf[NATIVE_HEIGHT][NATIVE_WIDTH];

// Our game boards.  One bit per cell. Allocate enough space
// for a scale of 1 (native resolution).
unsigned char boards[2][(NATIVE_WIDTH * NATIVE_HEIGHT / 8)];
unsigned char *curBoard;
unsigned char *prevBoard;

bool paused = false;

// Rules for calculating the next state of a cell.
// Living cells with 2 or 3 living neighbors stay alive. Otherwise, they die
// Dead cells with 3 living neighbors comt to live. Otherwise, they stay dead
bool aliveNext[9] = {false, false, true,  true, false, false, false, false, false};
bool deadNext[9]  = {false, false, false, true, false, false, false, false,  false};

// Check the value of an individual cell in a board (with wrap around)
bool boardRead(unsigned char *brd, int x, int y)
{
  int xMod = (x + curWidth) % curWidth;
  int yMod = (y + curHeight) % curHeight;
  int i = xMod + (yMod * curWidth);
  int j = i % 8;
  i = (i - j) / 8;
  return bitRead(brd[i], j);  
}

// Set the value of an individual cell in a board (with wrap around)
void boardWrite(unsigned char *brd, int x, int y, bool val)
{
  int xMod = (x + curWidth) % curWidth;
  int yMod = (y + curHeight) % curHeight;
  int i = xMod + (yMod * curWidth);
  int j = i % 8;
  i = (i - j) / 8;
  bitWrite(brd[i], j, val);  
}

// Initialize curBord to a random configuration
void initBoard()
{
  Serial.printf("Board size: %d x %d\n", curWidth, curHeight);
  for (int y = 0; y < curHeight; y++) {
    for (int x = 0; x < curWidth; x++) {
      boardWrite(curBoard, x, y, random(100) < PROB);
    }
  }  
}

// Calculate the next state of the board
void nextBoard()
{
  for (int y = 0; y < curHeight; y++) {
    for (int x = 0; x < curWidth; x++) {
      bool cur = boardRead(curBoard, x, y);
      // Count how many living neighbord the cell has
      int neighbors = boardRead(curBoard, x-1, y-1);
      neighbors +=    boardRead(curBoard, x,   y-1);
      neighbors +=    boardRead(curBoard, x+1, y-1);
      neighbors +=    boardRead(curBoard, x-1, y);
      neighbors +=    boardRead(curBoard, x+1, y);
      neighbors +=    boardRead(curBoard, x-1, y+1);
      neighbors +=    boardRead(curBoard, x,   y+1);
      neighbors +=    boardRead(curBoard, x+1, y+1);

      bool next = cur ? aliveNext[neighbors] : deadNext[neighbors];
      boardWrite(prevBoard, x, y, next);
    }
  }  

  // Swap the two boards
  unsigned char *bTemp = curBoard;
  curBoard = prevBoard;
  prevBoard = bTemp;
}

// Draw the current board to the screen
void drawBoard()
{
  //Serial.println("drawBoard");
  for (int y = 0; y < curHeight; y++)
  {
    for (int x = 0; x < curWidth; x++)
    {
      // Calculate the color to show based on the current and previous state of a cell.
      bool cur = boardRead(curBoard, x, y);
      bool prev = boardRead(prevBoard, x, y);
      unsigned short boardColor;
      if (cur && prev)
        boardColor = ALIVE;
      else if (cur && !prev)
        boardColor = BIRTH;
      else if (!cur && prev)
        boardColor = DEATH;
      else
        boardColor = DEAD;

      // Set the screen buffer (with scaling)
      for (int j = 0; j < scale; j++)
      {
        for (int i = 0; i < scale; i++)
        {
          screenBuf[(y * scale) + j][(x * scale) + i] = boardColor;
        }
      }
    }
  }
  tft.pushImage(0, 0,  NATIVE_WIDTH, NATIVE_HEIGHT, (unsigned short *)screenBuf);
}

//! Long time delay, it is recommended to use shallow sleep, which can effectively reduce the current consumption
void espDelay(int ms)
{   
    esp_sleep_enable_timer_wakeup(ms * 1000);
    esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_PERIPH,ESP_PD_OPTION_ON);
    esp_light_sleep_start();
}

// Setup our button handlers for the following:
//   Button 1 long click - cycle through the three scaling factors
//   Button 1 click -      reset the game board to a new random layout
//   Button 2 long click - pause or unpause the game
//   Button 2 click -      if paused, go to the next board state
void button_init()
{
    btn1.setLongClickHandler([](Button2 & b) {
        scale += 2;
        if (scale > 5)
          scale = 1;

        curWidth = NATIVE_WIDTH / scale;
        curHeight = NATIVE_HEIGHT / scale;
        initBoard();
        drawBoard();
    });

    btn1.setClickHandler([](Button2 & b) {
        initBoard();
        drawBoard();
    });

    btn2.setLongClickHandler([](Button2 & b) {
      paused = !paused;
    });

    btn2.setClickHandler([](Button2 & b) {
      if (paused)
      {
        nextBoard();
        drawBoard();
      }
    });
}

void button_loop()
{
    btn1.loop();
    btn2.loop();
}

void setup()
{
    Serial.begin(115200);
    Serial.println("Start");
    //randomSeed(analogRead(0));

    // Initialize the screen and turn on the backlight
    tft.init();
    tft.setRotation(0);
    tft.fillScreen(ALIVE);
    tft.setTextColor(DEAD);

    if (TFT_BL > 0) { // TFT_BL has been set in the TFT_eSPI library in the User Setup file TTGO_T_Display.h
         pinMode(TFT_BL, OUTPUT); // Set backlight pin to output mode
         digitalWrite(TFT_BL, TFT_BACKLIGHT_ON); // Turn backlight on. TFT_BACKLIGHT_ON has been set in the TFT_eSPI library in the User Setup file TTGO_T_Display.h
    }

    tft.setSwapBytes(true);

    // Initialize the game board and display our starting image
    curBoard = (unsigned char *)boards[0];
    prevBoard = (unsigned char *)boards[1];
    initBoard();
    drawBoard();

    button_init();
}

void loop()
{
  if (!paused)
  {
    // Compute the next board and display it
    //espDelay(200);
    nextBoard();
    button_loop();
    drawBoard();
  }
  button_loop();
}
